"""OpenAI payload handlers."""

from typing import Any

from .base import ModelPayloadHandler


class OpenAICompletionsPayloadHandler(ModelPayloadHandler):
    """Payload handler for OpenAI Chat Completions API."""

    def get_required_tool_choice(self) -> str | dict[str, Any]:
        """Get tool_choice value for OpenAI Completions API."""
        return "required"



"""Vertex Gemini payload handler."""

from typing import Any

from .base import ModelPayloadHandler


class VertexGeminiPayloadHandler(ModelPayloadHandler):
    """Payload handler for Google Vertex AI Gemini API."""

    def get_required_tool_choice(self) -> str | dict[str, Any]:
        """Get tool_choice value for Vertex Gemini API."""
        return "any"



"""Process tool creation for UiPath process execution."""

from typing import Any

from langchain_core.tools import StructuredTool
from langgraph.types import interrupt
from uipath.agent.models.agent import AgentIxpExtractionResourceConfig
from uipath.eval.mocks import mockable
from uipath.platform.attachments import Attachment
from uipath.platform.common import DocumentExtraction
from uipath.platform.documents import ExtractionResponseIXP

from .structured_tool_with_output_type import StructuredToolWithOutputType
from .utils import sanitize_tool_name


def create_ixp_extraction_tool(
    resource: AgentIxpExtractionResourceConfig,
) -> StructuredTool:
    """Uses interrupt() to suspend graph execution until data is extracted (handled by runtime)."""
    tool_name: str = sanitize_tool_name(resource.name)
    project_name = resource.properties.project_name
    version_tag = resource.properties.version_tag

    @mockable(
        name=resource.name,
        description=resource.description,
        input_schema=Attachment.model_json_schema(),
        output_schema=ExtractionResponseIXP.model_json_schema(),
        example_calls=resource.properties.example_calls,
    )
    async def extraction_tool_fn(**kwargs: Any) -> ExtractionResponseIXP:
        from uipath.platform import UiPath

        uipath = UiPath()

        attachment_id = kwargs.get("id")
        attachment_full_name = kwargs.get("full_name")

        # TODO: attachment_mime_type is currently not used anywhere (attachment_full_name will also be obsolete once attachments api is onboarded)
        # should we use them somewhere else? otherwise input_schema should only contain the file id
        # attachment_mime_type = kwargs.get("mime_type")

        # TODO: current workaround. DocumentExtraction model should support attachment_id and use the
        # start_ixp_extraction_from_attachment sdk method once support is added

        attachment_local_file_path = await uipath.attachments.download_async(
            key=attachment_id, destination_path=attachment_full_name
        )
        return interrupt(
            DocumentExtraction(
                project_name=project_name,
                tag=version_tag,
                file_path=attachment_local_file_path,
            )
        )

    tool = StructuredToolWithOutputType(
        name=tool_name,
        description=resource.description,
        args_schema=Attachment,
        coroutine=extraction_tool_fn,
        output_type=ExtractionResponseIXP,
    )

    return tool



"""OpenAI payload handlers."""

from typing import Any

from .base import ModelPayloadHandler


class OpenAIResponsesPayloadHandler(ModelPayloadHandler):
    """Payload handler for OpenAI Responses API."""

    def get_required_tool_choice(self) -> str | dict[str, Any]:
        """Get tool_choice value for OpenAI Responses API."""
        return "required"



"""Bedrock Invoke payload handler."""

from typing import Any

from .base import ModelPayloadHandler


class BedrockInvokePayloadHandler(ModelPayloadHandler):
    """Payload handler for AWS Bedrock Invoke API."""

    def get_required_tool_choice(self) -> str | dict[str, Any]:
        """Get tool_choice value for Bedrock Invoke API."""
        return {"type": "any"}



"""Dict-like object reducers for merging state with field-specific reducers."""

from typing import Any

from pydantic import BaseModel
from uipath.platform.attachments import Attachment


def add_job_attachments(
    left: dict[str, Attachment], right: dict[str, Attachment]
) -> dict[str, Attachment]:
    """Merge attachment dictionaries, with right values taking precedence.

    This reducer function merges two dictionaries of attachments by UUID string.
    If the same UUID exists in both dictionaries, the value from 'right' takes precedence.

    Args:
        left: Existing dictionary of attachments keyed by UUID string
        right: New dictionary of attachments to merge

    Returns:
        Merged dictionary with right values overriding left values for duplicate keys
    """
    if not right:
        return left

    if not left:
        return right

    return {**left, **right}


def merge_objects(left: Any, right: Any) -> Any:
    """Merge a Pydantic model with another model or dict, with right values taking precedence.

    Applies field-specific reducers from annotation metadata when merging values.

    Args:
        left: Existing Pydantic BaseModel instance
        right: New Pydantic BaseModel instance or dict to merge

    Returns:
        New Pydantic model instance with merged values

    Raises:
        TypeError: If left is not a Pydantic BaseModel or right is not a BaseModel or dict
    """
    if not right:
        return left

    if not left:
        return right

    # validate input types
    if not isinstance(left, BaseModel):
        raise TypeError("Left object must be a Pydantic BaseModel")

    if not isinstance(right, (BaseModel, dict)):
        raise TypeError("Right object must be a Pydantic BaseModel or dict")

    model_fields = type(left).model_fields
    merged_values = {}

    for field_name in model_fields:
        merged_values[field_name] = getattr(left, field_name)

    for field_name in model_fields:
        if isinstance(right, BaseModel):
            if hasattr(right, field_name):
                right_value = getattr(right, field_name)
            else:
                continue  # field not present in right
        else:
            # right is dict
            if field_name not in right:
                continue  # field not present in right
            right_value = right[field_name]

        field_info = model_fields[field_name]
        left_value = merged_values[field_name]

        # apply reducer if defined
        if field_info.metadata and callable(field_info.metadata[0]):
            reducer_func = field_info.metadata[0]
            merged_values[field_name] = reducer_func(left_value, right_value)
        else:
            merged_values[field_name] = right_value

    # return new model instance with merged values
    return type(left)(**merged_values)
